#
# autoqa library - config settings
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Will Woods <wwoods@redhat.com>

import os
import os.path
import ConfigParser

# Block size to use for reading/writing data files
BLOCKSIZE = 4096 # Sure why not.
# Scratch/tmp dir.
scratchdir = None
# Default directory for log files etc. Usually $scratchdir/log
logdir = None
# Default location for keyval data to be stored. Usually $scratchdir/keyval
keyvaldir = None
# Default location for tests to look for external data files / binaries
# (e.g. minimon, ks.cfg)
bindir = "."
# Default location for cached RPM files (e.g. downloaded from Koji).
# Usually $scratchdir/rpms
rpmdir = None
# Default location for the pretty log template
log_template = "/etc/autoqa/log_template.html"

def set_scratchdir(newscratchdir):
    '''sets config.scratchdir, config.logdir, config.keyvaldir and config.rpmdir
    and creates those dirs if needed'''
    global scratchdir, logdir, keyvaldir, rpmdir
    scratchdir = newscratchdir
    logdir = os.path.join(newscratchdir, "log")
    keyvaldir = os.path.join(scratchdir, "keyval")
    rpmdir = os.path.join(scratchdir, "rpms")
    for dir_ in [scratchdir, logdir, keyvaldir, rpmdir]:
        if not os.path.isdir(dir_):
            os.makedirs(dir_)
# Stupid default scratch dir.
# XXX TODO: use mktemp?
set_scratchdir("/tmp/autoqa")

class SingleConfigParser(ConfigParser.SafeConfigParser):
    '''This is a subclassed SafeConfigParser. It has several additional methods
       serving our needs.
    '''
    def read_single(self, filenames):
        '''This is the same as the read() method, but it parsers only the first
           available config file from the list, not all of them.
        '''
        if isinstance(filenames, str):
            filenames = [filenames]
        for f in filenames:
            retval = self.read(f)
            if retval:
                # file read, don't read any other
                return retval
        else:
            return []

    def get_section(self, section, default_conf={}):
        '''Read the whole section and convert it into a dictionary (with string
           values).

           Args:
           section -- section of the config to be retrieved
           default_conf -- default configuration values

           Returns:
           Dictionary containing retrieved data on success.

           Throws:
           ConfigParser.NoSectionError -- when no such section exists
        '''
        items = dict(self.items(section))
        retval = default_conf.copy()
        retval.update(items)
        return retval

def getbool(value):
    '''Converts a string into a boolean.
       Recognizes many usual strings: 1/0, yes/no, true/false, on/off; also
       empty strings.
       Throws ValueError if the string can't be recognized.
       Uses default bool conversion for non-string values.
    '''
    boolean_states = {'1': True, 'yes': True,  'true': True,   'on': True,
                      '0': False, 'no': False, 'false': False, 'off': False, '': False}

    if not isinstance(value, str):
        return bool(value)

    value = value.lower()
    if value not in boolean_states:
        raise ValueError, 'Not a boolean: %r %s' % (value, type(value))

    return boolean_states[value]


autoqa_conf = SingleConfigParser()
autoqa_conf.read('/etc/autoqa/autoqa.conf')

if __name__ == '__main__':
    print "logdir = %s" % logdir
