#!/usr/bin/python
#
# watch-koji-builds.py - script to watch a set of koji tags for new builds
#                        and kick off tests when new builds/packages
#                        are found.
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#     Josef Skladanka <jskladan@redhat.com>

import os
import sys
import time, calendar
import pickle
import subprocess
import optparse
from autoqa import koji_utils
from autoqa import bodhi_utils
from autoqa.repoinfo import repoinfo
from autoqa.util import get_basearch
from autoqa.filelock import FileLock

class KojiWatcher(object):
    kojiserver = 'http://koji.fedoraproject.org/kojihub'
    kojiopts = {} # Possible items: user, password, debug_xmlrpc, debug..
    cachedir = '/var/cache/autoqa/watch-koji-builds/'
    cachefile = os.path.join(cachedir, 'koji_history_prevtimes.cache')

    def __init__(self, verbose = False, dry_run = False, prevtime = None, only_tags = []):
        self.dry_run = dry_run
        self.verbose = verbose

        if self.verbose:
            print "Connecting to koji:", self.kojiserver
        self.session = koji_utils.SimpleKojiClientSession(self.kojiserver, self.kojiopts)
        self.session.check_connection()

        # If user did not set the -prevtime parameter, use cached
        if prevtime is None:
            if self.verbose:
                print "Loading prevtimes from cache"
            self.default_prevtime = calendar.timegm(time.gmtime()) - 3*60*60 # now - 3 hours in UTC
            self.prevtimes = self._load_prevtimes_from_cache() # returns dict
            # self.prevtimes is a dict. if the respective key is not found,
            # while checking the 'time of the last update', default_prevtime
            # is used instead
        # but if the prevtime was set, we must make sure to use it.
        else:
            self.default_prevtime = prevtime
            self.prevtimes = {} # if this is empty, default_prevtime is used (see list_new_builds)
        if self.verbose:
            print "Loading taglist"
        self.taglist, self.master_repos, self.pending_repos = self._load_repoinfo(only_tags) # returns tuple of sets

    def _load_repoinfo(self, only_tags=[]):
        """
        Loads the tags from repoinfo.conf using autoqa.repoinfo.  If only_tags
        is provided, only the tags provided will be used.

        returns (taglist, master_repos, pending_repos)
            taglist        =   Set of all tags found in repoinfo
            master_repos   =   Set of repos which do not have 'parent' i.e. the 'main' or 'stable' repos.
                                Master repos do not have appropriate '-pending' tag in koji, because
                                no new packages get to these repos.
                                These are internaly used for grouping & filtering of builds.
            pending_repos =    Set of all '-pending' tags in koji.

        """

        def is_tag_requested(tag, only_tags):
            '''
                if only_tags is empty, returns True
                If only_tags is non-empty, returns whether the provided tag is
                a member of only_tags.
            '''
            if not only_tags:
                return True
            return tag in only_tags

        taglist = set()
        master_repos = set()
        pending_repos = set()
        for repo in repoinfo.repos():
            # Always include the tag for the development branch (aka rawhide)
            if repoinfo.get(repo, "collection_name") == "devel":
                tagname = repoinfo.get(repo, "tag")
                if is_tag_requested(tagname, only_tags):
                    taglist.add(tagname)
            else:
                '''
                for all the repos do:
                    if repo has no parent (i.e. is base repo):
                        add '$repo$-updates-candidate' to taglist
                    else:
                        add '$repo$-pending' to taglist

                NOTE: This assumes that updates are tagged in the format 'dist-fNN-updates-candidate'
                '''
                if len(repoinfo.getparents(repo)) == 0:
                    tagname = repoinfo.get(repo, "tag") + "-updates-candidate"
                    if is_tag_requested(tagname, only_tags):
                        taglist.add(tagname)
                        master_repos.add(repoinfo.get(repo, "tag")) #save the 'master' repo for later filtering
                else:
                    tagname = repoinfo.get(repo, "tag") + "-pending"
                    if is_tag_requested(tagname, only_tags):
                        taglist.add(tagname)
                        pending_repos.add(tagname)

        if self.verbose:
           print "\ntaglist:", sorted(taglist), "\n\n", "pending_repos:", sorted(pending_repos), "\n"
        return taglist, master_repos, pending_repos

    def _load_prevtimes_from_cache(self):
        """
        Returns stored prevtimes - i.e. the create_ts of the last package in the appropriate tag.
        Return format = {tag : prevtime, ...}
        """
        try:
            f = open(self.cachefile)
            prevtimes = pickle.load(f)
            f.close()
            return prevtimes
        except:
            return {}

    def _save_prevtimes_to_cache(self):
        """
        Saves self.prevtimes to self.cachefile.
        This cache is used to determine the from which to continue during next runs.
        """
        try:
            os.makedirs(self.cachedir)
        except OSError, e:
            if e.errno != 17: # already exists
                raise
        f = open(self.cachefile, "w")
        pickle.dump(self.prevtimes, f)
        f.close()

    def new_builds_since(self, tag, prevtime):
        """
        Inspects the tag history in Koji, and returns all new "active" packages for that tag
        which have create_ts > prevtime.

        Internaly used only for -pending tags, because this tends to be sloooow for
        -updates-candidate.

        Return format = [koji_build_info, ...]
        """
        updates = []
        # get history from koji & filter everything newer than prevtime
        if self.verbose:
            print "    Fetching tagHistory from koji"
        updates = filter(lambda pkg: pkg['create_ts'] > prevtime, self.session.tagHistory(tag = tag))
        if self.verbose:
            print "    Sorting according to create_ts"
        updates.sort(key = lambda pkg: pkg['create_ts'])
        # sanity_check - filter just "active" builds
        if self.verbose:
            print "    Filtering out non-active builds"
        non_active = filter(lambda pkg: pkg['active'] is None, updates)
        #TODO: print non-active builds to stderr - these skipped our attention somehow
        if self.verbose:
            print "    Filtering active builds"
        updates = filter(lambda pkg: pkg['active'] is not None, updates)

        # update the appropriate prevtime in self.prevtimes
        if len(updates) > 0:
            self.prevtimes[tag] = updates[-1]['create_ts']

        # sadly the tagHistory does not all the required information
        # we need to run the tests, so let's fetch the 'good' info
        if self.verbose:
            print "    Getting real info about builds"
        for i in range(len(updates)):
            b = updates[i]
            updates[i] = self.session.getBuild("%s-%s-%s" % (b['name'], b['version'], b['release']))
            # TODO: maybe this is bug in KOJI - check it out
            # because getBuild returns 'id' instead of 'build_id'
            updates[i]['build_id'] = updates[i]['id']

        return updates

    def new_builds_in_updates_candidate(self, taglist, prevtime):
        """
        Because listing tag history for -updates-candidate is slooow,
        we're internally using this method to deal with updates-candidate tags.
        This is the 'old' implementation from post-koji event

        Hopefully koji will get an enhancement, which will allow to query
        only a subset of the tag history, so it would be significantly faster.

        This is now being consulted with jkeating.

        Return format = {tag: [new_builds], ...}
        """

        untagged_builds = []
        tagged_builds = {}
        for tag in taglist:
            tagged_builds[tag] = []

        if self.verbose:
            print "    Fetching builds since:", prevtime

        builds = self.session.list_builds_since(prevtime)
        try:
            builds += self.prevtimes['untagged_builds'] # untagged builds from previous run
            if self.verbose:
                print "    Adding previously stored untagged builds"
        except KeyError:
            # no previously stored undagged builds, pass
            if self.verbose:
                print "    No previously stored untagged builds"
            pass

        if self.verbose:
            print "    Sorting according to completion_ts"
        builds.sort(key = lambda pkg: pkg['completion_ts'])
        try:
            self.prevtimes['updates_candidate'] = builds[-1]['completion_ts']
        except IndexError:
            if self.verbose:
                print "    No new builds since last check"
            pass

        if self.verbose:
            print "    Assorting according to tags"
        for nvr in builds:
            tags = set(self.session.tag_history(nvr))
            if tags:
                # Only check packages with tags from the taglist
                for t in taglist.intersection(tags):
                    tagged_builds[t].append(nvr)
            elif not self.session.build_failed(nvr):
                # No tag yet but the build is still OK - check again later
                untagged_builds.append(nvr)
        if self.verbose:
            print "    Assorting done"

        self.prevtimes['untagged_builds'] = untagged_builds

        return tagged_builds

    def list_new_builds(self):
        """
        Cretes a list of new builds for all the tags specified in repoinfo.conf
        (see _load_repoinfo for more information). It also updates the self.prevtimes
        to the 'create_ts' of the build in the appropriate tag.

        There are more tags for each distro:
          '*-updates-candidate' which contains
             1) all updates which have not yet been pushed to the '-testing' or '-updates'
                and which are not yet in '-pending'
             2) all updates which have '-pending' tag.

        Because '-pending' have also the '-updates-candidate' for the same build,
        we remove all the builds which are in '-pending' from the '-updates-candidate'
        (so we do not have duplicate entries).

        Return format: {tag: [list of packages], ...}
        """
        history = {}
        # this will list the tag history from koji for each tag
        # we need to care just about the still 'active' builds
        # so let's filter these, and store the list of 'active' packages
        # into the dictionary with the tag as key
        # so we'll have something like:
        #   history['dist-f14-updates-pending'] = [list of active packages in f14-updates-pending]
        #   ...
        taglist = list(self.taglist)
        for tag in filter(lambda x: x.endswith("-pending"), taglist):
            if self.verbose:
                print "Checking", tag
            try:
                prevtime = self.prevtimes[tag]
            except KeyError:
                prevtime = self.default_prevtime

            if self.verbose:
                print 'Searching for new builds since %s UTC' % time.asctime(time.gmtime(prevtime))
            history[tag] = self.new_builds_since(tag, prevtime)

        # checking the non-pending tags the old way
        taglist = filter(lambda x: not x.endswith("-pending"), taglist)
        if self.verbose:
            print "Checking the old way:", sorted(taglist)

        try:
            prevtime = self.prevtimes['updates_candidate']
        except KeyError:
            prevtime = self.default_prevtime

        if self.verbose:
                print 'Searching for new builds since %s UTC' % time.asctime(time.gmtime(prevtime))
        # fetch the new builds
        new_builds = self.new_builds_in_updates_candidate(set(taglist), prevtime)
        # and store them to the history dict the 'right' way
        for tag in new_builds.keys():
            history[tag] = new_builds[tag]

        # let's return the dictionary for further usage
        return history


    def schedule_jobs(self, new_builds):
        """
        This should actually do the appropriate scheduling -
        i.e. calling the harness in the required way.

        This should be the only method to reimplement in all
        the different post-koji wathers (i.e. the batch watcher etc.)

        Or we can implement several 'schedule_jobs' methods in
        this particular instance, and run them in the run() method.
        """

        # Original post-koji build implementation
        exit_status = 0
        # schedule koji events
        for tag in sorted(new_builds.keys()):
            bodhi_skip = []
            if self.verbose:
                print "Scheduling tests for", tag
            for b in new_builds[tag]:
                # Get a list of all package arches in this build
                arches = [r['arch'] for r in self.session.listRPMs(b['build_id'])]
                arches = map(get_basearch, arches)
                harnesscall = ['autoqa']
                p_tag = tag.replace('-pending', '')
                bodhi_info = None
                stashed_nvrs = []

                # -pending denotes bodhi update event
                if tag.endswith('-pending'):
                    # because bodhi update can contain multiple nvrs (the 'title')
                    # we need a way to skip those 'redundant' nvrs.
                    # for each build, we query bodhi for the info, list
                    # nvrs in that bodhi update, and if multiple nvrs are found
                    # they are added to the 'bodhi_skip' list, so we won't schedule
                    # the autoqa bodhi-update events for this particular nvrs

                    if b['nvr'] in bodhi_skip:
                        if self.verbose:
                            print "Skipping %s since it's a part of already scheduled bodhi update" % b['nvr']
                        continue

                    bodhi_info = bodhi_utils.query_update(b['nvr'])

                    if bodhi_info is None:
                        print >> sys.stderr, "Update %s not found in Bodhi" % b['nvr']
                        continue

                    # inspect the 'builds' part of bodhi info.
                    # if multiple builds are found, add all the respective
                    # nvrs to the skip list, so we won't schedule reciprocial
                    # bodhi update events.
                    if len(bodhi_info['builds']) > 1:
                        for to_skip in bodhi_info['builds']:
                            bodhi_skip.append(to_skip['nvr'])
                            stashed_nvrs.append(to_skip['nvr'])
                        stashed_nvrs.remove(b['nvr'])

                    harnesscall += ['post-bodhi-update', '--title', bodhi_info['title'],
                                   '--targettag', p_tag]
                    if bodhi_info['updateid']:
                        harnesscall += ['--updateid', bodhi_info['updateid']]

                else:
                    harnesscall += ['post-koji-build', '--kojitag', tag]

                # Invoke autoqa with an --arch arg for each arch we care about
                repoarches = set(repoinfo.getrepo_by_tag(p_tag).get("arches"))
                testarches = set(repoarches).intersection(arches)

                for arch in testarches:
                    harnesscall += ['--arch', arch]
                nvr = koji_utils.getENVR(b)
                harnesscall.append(nvr)

                # if some builds were skipped during bodhi-event planning
                # add the nvrs to the respective call
                harnesscall.extend(stashed_nvrs)

                if self.dry_run or self.verbose:
                    print " ".join(harnesscall)

                # do not schedule jobs on dry_run
                if self.dry_run:
                    continue

                retval = subprocess.call(harnesscall)
                if retval != 0:
                    exit_status = 10

        return exit_status

    def schedule_jobs_batch(self, new_builds):
        """
        This is alternate job scheduler for the 'virtual'
        post-koji-build-batch watcher.

        We decided, that pulling the information from koji twice
        would not be efficient, so we schedule both 'per package' and
        'batch' jobs using the data gathered once.
        """
        exit_status = 0
        if self.verbose:
                print "Scheduling batch tests"
        for tag in sorted(new_builds.keys()):
            # Skip tags with no new entries
            if len(new_builds[tag]) == 0:
                continue
            harness_arches = set()
            harness_nvrs = []
            p_tag = tag.replace('-pending', '')
            repoarches = set(repoinfo.getrepo_by_tag(p_tag).get("arches"))

            for b in new_builds[tag]:
                # do not query koji for arches, if we already have rpms from all
                # possible arches from the repoarches
                if harness_arches != repoarches:
                    arches = [r['arch'] for r in self.session.listRPMs(b['build_id'])]
                    arches = map(get_basearch, arches)
                    testarches = set(repoarches).intersection(arches)
                    harness_arches.update(testarches)

                nvr = koji_utils.getENVR(b)
                harness_nvrs.append(nvr)

            harnesscall = ['autoqa']
            # bodhi events
            if tag.endswith('-pending'):
                harnesscall += ['post-bodhi-update-batch', '--targettag', p_tag]
            # koji events
            else:
                harnesscall += ['post-koji-build-batch', '--kojitag', tag]

            for a in harness_arches:
                harnesscall += ['--arch', a]

            harnesscall.extend(harness_nvrs)

            if self.dry_run or self.verbose:
                print " ".join(harnesscall)
            # do not schedule jobs on dry_run
            if self.dry_run:
                continue

            retval = subprocess.call(harnesscall)
            if retval != 0:
                exit_status = 10

        return exit_status


    def run(self):
        """
        Call this method to run the watcher.

        It calls all the method in order to schedule jobs to autotest.
        (or list them, if self.dry_run is set to true)
        """


        if self.verbose:
            print "Looking for new builds"
        new_builds = self.list_new_builds() # this updates self.prevtimes
        if not self.dry_run:
            if self.verbose:
                print "Saving prevtimes to cache"
            self._save_prevtimes_to_cache()

        if self.verbose:
            print "Scheduling jobs"

        exit_status = []
        exit_status.append(self.schedule_jobs(new_builds))
        exit_status.append(self.schedule_jobs_batch(new_builds))
        for e in exit_status:
            if e != 0:
                return e
        return 0



if __name__ == '__main__':
    # Set up the option parser
    # TODO: optparse for prevtime
    parser = optparse.OptionParser(description='Script to watch a set of koji \
tags for new builds and kick off tests when new builds/packages are found.')
    parser.add_option('--dryrun', '--dry-run', action='store_true',
        help='Do not actually execute commands, just show what would be done')
    parser.add_option('-t', '--tag', action='append',
        type='string', default=[], dest="tags",
        help='Only look for builds in the specified koji tags (default: all)')
    parser.add_option('-p', '--prevtime', action='store',
        type='float', default=None,
        help='How far back to look for new builds (seconds since last epoch)')
    parser.add_option('--verbose', action='store_true',
        help='Print extra information')
    (opts, args) = parser.parse_args()


    # Because paralel runs of koji watcher caused problems with not-updated
    # cache file (prevtimes), we use locking mechanism to prevent paralel runs.
    #
    # The problem was:
    #   1) new KojiWatcher instance (loads data from cachefile in __init__)
    #   2) watcher runs (queries koji/bodhi - can take long time)
    #   3) after all the querying, new data is saved to the cachefile
    #
    # If new watcher runs during the step (2), then it uses the same
    # cached data as the 'already running' instance, which emerges in two
    # (or more) jobs scheduled for the same update/test combination.
    #
    # Lock is not created on dry-run, since dry-run does not update cachefile.

    lock = FileLock(KojiWatcher.cachefile)
    if opts.dryrun is not True:
        if lock.acquire() is not True:
            print "Another instance of this watcher is already running. Exiting..."
            sys.exit(0)

    try:
        watcher = KojiWatcher(verbose = opts.verbose, dry_run = opts.dryrun, prevtime = opts.prevtime, only_tags = opts.tags)
        exit_status = watcher.run()

    except KeyboardInterrupt:
        print "Exiting on keyboard interrupt."
        lock.release()
        sys.exit(1)

    lock.release()
    sys.exit(exit_status)
