#!/usr/bin/python
#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Chris Lumens <clumens@redhat.com>

import sys

from cases import *

class AutopartRunner(BaseRunner):
    def writeKickstart(self):
        fo = open("/tmp/kickstart.ks", "w")
        fo.write("autopart\n")
        fo.write("clearpart --all\n")
        fo.write("zerombr")
        fo.close()

class AutopartTestCase(BaseTestCase):
    """What autopartitioning made is dependent upon how much disk space is
       in the VM:

       < 50GB:  /boot; LVM containing / and swap
       >= 50GB: /boot; LVM containing /, /home, and swap

       This test is not yet machine-specific, but could be made so.  That would
       involve testing for the presence of /boot/EFI or PReP, for instance.
    """

    def test_boot(self):
        self.default_test_boot()

    def test_home(self):
        size = 0
        for d in self.disks(self.state):
            size += d["size"]

        # This test only runs if there's enough disk space for /home to have
        # been created.
        if size < 50*1024:
            return True

        # Check that the size of / is no more than 50 GB.
        self.assert_(rootDev["size"] < 50*1024, "/ is too big.")

        homeDev = self.single(self.getDevs(self.state, mountpoint="/home"))
        self.failIf(not homeDev,
                    "autopart did not create /home.")

        # Check the fs type and underlying device of /home.
        self.assert_(homeDev["format"]["type"] == "ext4",
                     "Type of /home is not ext4.")
        self.assert_(homeDev["type"] == "lvmlv",
                     "/home was not created on a logical volume.")

    def test_root(self):
        rootDev = self.single(self.getDevs(self.state, mountpoint="/"))
        self.failIf(not rootDev,
                    "autopart did not create /.")

        # Check the fs type and underlying device of /.
        self.assert_(rootDev["format"]["type"] == "ext4",
                     "Type of / is not ext4.")
        self.assert_(rootDev["type"] == "lvmlv",
                     "/ was not created on a logical volume.")

    def test_swap(self):
        from pyanaconda.iutil import swapSuggestion

        swapDev = self.single(self.getDevs(self.state, format="swap"))
        self.failIf(not swapDev,
                    "autopart did not create swap.")

        # Swap size can vary wildly, so just check that swap exists and is inside
        # the range anaconda suggests.
        (minSwap, maxSwap) = swapSuggestion()
        self.assert_(minSwap <= swapDev["size"] <= maxSwap,
                     "swap is not within suggested range")

        # Check the fs type and underlying device of swap.
        self.assert_(swapDev["format"]["type"] == "swap",
                     "Type of swap is not swap.")
        self.assert_(swapDev["type"] == "lvmlv",
                     "swap was not created on a logical volume.")
